// Copyright (c) 2016, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

part of passive_data;

const _primitive = const ['num', 'int', 'double', 'bool', 'String'];
const _primitiveDynamic = const [
  'dynamic',
  'num',
  'int',
  'double',
  'bool',
  'String',
  'List'
];

/// Test if the given type String is a passive data structure.
bool isPdsType(String type) {
  type = type.replaceAll(' ', '');
  if (_primitiveDynamic.contains(type)) {
    return true;
  } else {
    if (type.startsWith('List<')) {
      return isPdsType(type.substring(5, type.length - 1));
    } else if (type.startsWith('Map<String,')) {
      return isPdsType(type.substring(11, type.length - 1));
    } else {
      return false;
    }
  }
}

/// Create PDS convert code for the given [type] where [data] is used to access
/// the object that contains the original data. The generated code depends on
/// the `passive_data` library.
String generateTypeToPdsCode(String type, String data) {
  if (isPdsType(type)) {
    return '$data';
  } else if (type == 'List' || type.startsWith('List<')) {
    return 'pdsFromList($data)';
  } else if (type == 'Set' || type.startsWith('Set<')) {
    return 'pdsFromSet($data)';
  } else if (type.startsWith('Map<String,')) {
    return 'pdsFromMap($data)';
  } else if (type.startsWith('Map<int,')) {
    return 'pdsFromMap($data)';
  } else {
    return '$data.toPds()';
  }
}

/// Create reverse construction code for the given [type] where [data] is used
/// to access the object that contains the data to construct this type and
/// [pdsType] represents the PDS representation of [type]. [depth] represents
/// the recursive level. This is required to name the List generator iteration
/// index variable. If [cast] is true, primitive types are also casted
/// (sometimes this is not required due to an encapsulating type check)
String generatePdsToTypeCode(String type, String data,
    [String pdsType = '', bool castInput = true, bool castOutput = true]) {
  // Strip off all spaces.
  type = type.replaceAll(' ', '');
  pdsType = pdsType.replaceAll(' ', '');

  // pdsType is either empty or complementary.
  if (pdsType.isNotEmpty && !isComplementaryPdsType(type, pdsType)) {
    throw new ArgumentError.value(
        pdsType, 'pdsType', 'must be either empty or complementary to type');
  }

  // If pdsType is empty and type is already a PDS type, you can assign type
  // to pdsType.
  if (pdsType.isEmpty && isPdsType(type)) {
    pdsType = type;
  }

  return _generatePdsToTypeCode(type, pdsType, data, castInput, castOutput);
}

/// Inner function of getPdsToTypeCode
String _generatePdsToTypeCode(
    String type, String pdsType, String data, bool castInput, bool castOutput) {
  if (type == 'dynamic') {
    return data;
  } else if (_primitiveDynamic.contains(type)) {
    return castOutput ? '$data as $type' : data;
  } else if (type.startsWith('List<') || type.startsWith('Set<')) {
    // Lists and Sets are reconstucted using their .from or .generate
    // constructor.
    var innerType = _unwrapListSetType(type);
    bool isSet = type.startsWith('Set<');

    // When primitive type parameters are used, the .from constructor can be
    // used.
    if (_primitiveDynamic.contains(innerType)) {
      return 'new $type.from($data)';
    } else {
      // At this point we require a complementary PDS type.
      if (pdsType.isEmpty) {
        throw new ArgumentError.value(
            pdsType, 'pdsType', 'must contain a complementary type');
      }

      // Use pdsToSet or pdsToList
      var innerPdsType = _unwrapListSetType(pdsType);
      var innerConstructor =
          _generatePdsToTypeCode(innerType, innerPdsType, 'item', false, false);

      // Do not include convert parameter if the innerConstructor doesn't do
      // anything.
      //
      // Note that the input is already casted in the closure.
      var convertParam = innerConstructor != 'item'
          ? ', ($innerPdsType item) => $innerConstructor'
          : '';

      var inputData = castInput ? '$data as List' : '$data';
      return (isSet
              ? 'pdsToSet($inputData$convertParam)'
              : 'pdsToList($inputData$convertParam)') +
          (castOutput ? ' as $type' : '');
    }
  } else if (type.startsWith('Map<String') || type.startsWith('Map<int')) {
    // Maps are regenerated using its .from or .fromIterable constructor.
    var innerType = _unwrapMapStringIntType(type);
    var intMap = type.startsWith('Map<int,');

    // When primitive type parameters are used, and the target Map is also
    // String indexed, the .from constructor can be used.
    if (!intMap && _primitiveDynamic.contains(innerType)) {
      return 'new $type.from($data)';
    } else {
      // At this point we require a complementary PDS type.
      if (pdsType.isEmpty) {
        throw new ArgumentError.value(
            pdsType, 'pdsType', 'must contain a complementary type');
      }

      // Use pdsToStringMap or pdsToIntMap
      //
      // Note that the input is already casted in the closure.
      var innerPdsType = _unwrapMapStringIntType(pdsType);
      var innerConstructor =
          _generatePdsToTypeCode(innerType, innerPdsType, 'item', false, false);

      // Do not include convert parameter if the innerConstructor doesn't do
      // anything.
      var convertParam = innerConstructor != 'item'
          ? ', ($innerPdsType item) => $innerConstructor'
          : '';

      var inputData = castInput ? '$data as Map<String, dynamic>' : '$data';
      return (intMap
              ? 'pdsToIntMap($inputData$convertParam)'
              : 'pdsToStringMap($inputData$convertParam)') +
          (castOutput ? ' as $type' : '');
    }
  } else {
    // Fallback to .fromPds constructor.
    // Note that fromPds parameters are usually already casted by their
    // encapsulating closure.
    var innerConstructor = _generatePdsToTypeCode(
        pdsType, pdsType, '$data', castInput, castOutput);
    return 'new $type.fromPds($innerConstructor)';
  }
}

/// Test if the [pdsType] (a PDS type) is complemenary to the Dart [type].
bool isComplementaryPdsType(String type, String pdsType) {
  if (!isPdsType(pdsType)) {
    throw new ArgumentError.value(pdsType, 'pdsType', 'must be a PDS type');
  }

  // Remove whitespaces to make life easier.
  type = type.replaceAll(' ', '');
  pdsType = pdsType.replaceAll(' ', '');

  // Recursively unwrap. The only difference should be in the PDS types for
  // non-PDS classes in [type].
  return _isComplementaryPdsType(type, pdsType);
}

/// Internal recursive function for isComplementaryPdsType
bool _isComplementaryPdsType(String type, String pdsType) {
  // 1. Unwrap type into primitive, List/Set/Map, or other class.
  // 2. Primitive? Should be equal to PDS type.
  // 3. List/Set/Map? Compare inner and outer type.
  // 4. Other? PDS type can be anything.

  const complementaryStart = const {
    'List<': 'List<',
    'Set<': 'List<',
    'Map<String,': 'Map<String,',
    'Map<int,': 'Map<String,'
  };

  if (_primitiveDynamic.contains(type)) {
    return type == pdsType;
  } else {
    for (var start in complementaryStart.keys) {
      if (type.startsWith(start)) {
        return pdsType.startsWith(complementaryStart[start]) &&
            _isComplementaryPdsType(
                type.substring(start.length, type.length - 1),
                pdsType.substring(
                    complementaryStart[start].length, pdsType.length - 1));
      }
    }

    // Not yet returned? PDS can be anything!
    return true;
  }
}

/// Get List/Set type parameter.
String _unwrapListSetType(String type) {
  return type.startsWith('List<')
      ? type.substring(5, type.length - 1)
      : type.substring(4, type.length - 1);
}

/// Get Map<String/int type parameter.
String _unwrapMapStringIntType(String type) {
  return type.startsWith('Map<String,')
      ? type.substring(11, type.length - 1)
      : type.substring(8, type.length - 1);
}
